/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2002 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Thanks to the Imlib2 authors and to Foley, van Dam, Feiner and
 * Hughe who wrote: "Computer Graphics: Principles and Practices"
 * Without them this new graphic toolkit would have been much harder!
 */

/* Short Description:
 *   An OO, anti-aliased and alpha-blending graphics primitives implementation.
 */

#ifndef GSMP_PIXMAP
#define GSMP_PIXMAP

// GSML stuff:

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh" // only for limit_min_max

// stl containers
#include <vector>
#include <deque>

// gtk--
#include <gtk--/drawingarea.h>

// FreeType2
#include <ft2build.h>
#include FT_FREETYPE_H

namespace GSMP
{
  template <class T>
  class _point {
  public:
    typedef T coord;
    
    _point () {
      x = y = 0;
    }
    
    _point (coord i_x, coord i_y) {
      x = i_x; y = i_y;
    }
    
    coord x, y;
  };
  
  template <class T>
  std::ostream& operator << (std::ostream& strm, const _point<T>& a) {
    strm << "[x: " << a.x << ", y: " << a.y << "]";
    return strm;
  }
  
  // ***
  
  template <class T>
  class _area {
  public:
    
    typedef T coord;
    typedef _point<coord> point;
    
    _area () {
      x = y = w = h = 0;
    }
    
    _area (coord i_x, coord i_y, coord i_w, coord i_h) {
      x = i_x; y = i_y; w = i_w; h = i_h;
    }
    
    void Clip (const _area<T>& a) {
      coord t_x = GSML::limit_min_max (x, a.x, a.x + a.w);
      coord t_y = GSML::limit_min_max (y, a.y, a.y + a.h);
      coord t_x2 = GSML::limit_min_max (x + w, a.x, a.x + a.w);
      coord t_y2 = GSML::limit_min_max (y + h, a.y, a.y + a.h);
      
      x = t_x;
      y = t_y;
      w = t_x2 - t_x;
      h = t_y2 - t_y;
    }
    
    void Extend (const _area<T>& a) {
      coord t_x = std::min (x, a.x);
      coord t_y = std::min (y, a.y);
      coord t_x2 = std::max (x + w, a.x + a.w);
      coord t_y2 = std::max (y + h, a.y + a.h);
      
      x = t_x;
      y = t_y;
      w = t_x2 - t_x;
      h = t_y2 - t_y;
    }
    
    bool Contains (const point& p) {
      return  (p.x >= x && p.x < x + w &&
	       p.y >= y && p.y < y + h);
    }
    
    bool Overlaps (const _area<T>& a) const {
      if (x < a.x)
	return a.x < x + w && a.y >= y && a.y < y + h;
      else
	return x < a.x + a.w && y >= a.y && y < a.y + a.h;
    }
    
    coord Value () const {
      return w * h;
    }
    
    bool operator== (const _area<T>& a) const {
      return x == a.x && y == a.y && w == a.w && h == a.h;
    }

    bool operator!= (const _area<T>& a) const {
      return ! operator== (a);
    }
    
    coord x, y, w, h;
  };
  
  template <class T>
  std::ostream& operator << (std::ostream& strm, const _area<T>& a) {
    strm << "[x: " << a.x << ", y: " << a.y << ", w: " << a.w << ", h: " << a.h << "]";
    return strm;
  }
  
  // ***
  
  class GsmpColor
  {
  public:
    
    typedef u_int8_t Component;
    typedef u_int32_t Color;
    
    static inline Component Red (const Color& color) {
      return (color >> 0) & 0xff;
    }
    static inline Component Green (const Color& color) {
      return (color >> 8) & 0xff;
    }
    static inline Component Blue (const Color& color) {
      return (color >> 16) & 0xff;
    }
    static inline Component Alpha (const Color& color) {
      return (color >> 24) & 0xff;
    }
    
    static void Rgba (const Color& color, Component* red, Component* green,
		      Component* blue, Component* alpha) {
      *red = (color >> 0) & 0xff;
      *green = (color >> 8) & 0xff;
      *blue = (color >> 16) & 0xff;
      *alpha = (color >> 24) & 0xff;
    }
    
    static void SetRgba (Color* color, const Color& value) {
      *color = value;
    }
    
    static void SetRgba (Color* color, Component red, Component green,
			 Component blue, Component alpha) {
      *color = alpha << 24 | blue << 16 | green << 8 | red;
    }
    
    // set/get some more color models (HVS, ...)
    
    // I know this code is suboptimal, but accessing unaligned memory
    // is opmized by the compiler.
    static inline void Blend (Color* color, const Color value) {
      int alpha = Alpha (value);
      if (alpha == 0xff) {
	*color = value;
      }
      else if (alpha != 0) {
	int tmp;
	
	tmp = (Red (value) - Red (*color) ) * alpha;
	int r = ((tmp + (tmp >> 8) + 128) >> 8) + Red (*color);
	
	tmp = (Green (value) - Green (*color) ) * alpha;
	int g = ((tmp + (tmp >> 8) + 128) >> 8) + Green (*color);
	
	tmp = (Blue (value) - Blue (*color) ) * alpha;
	int b = ((tmp + (tmp >> 8) + 128) >> 8) + Blue (*color);
	
	// TODO: think about alpha blending the alpha channel
	tmp = (Alpha (value) - Alpha (*color) ) * alpha;
	int a = ((tmp + (tmp >> 8) + 128) >> 8) + alpha;
	
	SetRgba (color, r, g, b, a);
      }
    }
    
    static inline void Blend (Color* color, const Color value,
			      const Component alpha_value) {
      int alpha = (Alpha (value) * alpha_value) >> 8;
      if (alpha == 0xff) {
	*color = value;
      }
      else if (alpha != 0) {
	int tmp;
	
	tmp = (Red (value) - Red (*color) ) * alpha;
	int r = ((tmp + (tmp >> 8) + 128) >> 8) + Red (*color);
	
	tmp = (Green (value) - Green (*color) ) * alpha;
	int g = ((tmp + (tmp >> 8) + 128) >> 8) + Green (*color);
	
	tmp = (Blue (value) - Blue (*color) ) * alpha;
	int b = ((tmp + (tmp >> 8) + 128) >> 8) + Blue (*color);
	
	// TODO: think about alpha blending the alpha channel
	tmp = (Alpha (value) - Alpha (*color) ) * alpha;
	int a = ((tmp + (tmp >> 8) + 128) >> 8) + alpha;
	
	SetRgba (color, r, g, b, a);
      }
    }
  };
  
  // ***
  
  class GsmpPixmap
  {
  public:
    typedef int coord;
    typedef _area<coord> area;
    typedef area::point point;
    
    GsmpPixmap ();
    GsmpPixmap (coord width, coord height);
    GsmpPixmap (const GsmpPixmap& source);
    
    ~GsmpPixmap ();
    
    // property and modifiers
    
    point Size () const;
    const area& Area () const;
    void Resize (const point& n_size);
    
    // operations
    
    void Clear (GsmpColor::Color c);
    
    void BitBlt (area source,  const point& dest_point);
    void BitBlt (const GsmpPixmap& pixmap, area source,  const point& dest_point);
    
    void DrawLine (point p1, point p2, GsmpColor::Color c);
    void DrawAaLine (point p1, point p2, GsmpColor::Color c);
    
    void DrawRectangle (const area& a, GsmpColor::Color c);
    void DrawFilledRectangle (const area& a, GsmpColor::Color c);
    
    void DrawCircle (const point& p1, coord r, GsmpColor::Color c);
    void DrawFilledCircle (const point& p1, coord r, GsmpColor::Color c);
    void DrawCircle (const point& p1, coord r, coord angle1, coord angle2,
		     GsmpColor::Color c);
    void DrawFilledCircle (const point& p1, coord r, coord angle1, coord angle2,
			   GsmpColor::Color c);
    
    void DrawAaCircle (const point& p1, coord r, GsmpColor::Color c);
    void DrawAaCircle (const point& p1, coord r, coord angle1, coord angle2,
		       GsmpColor::Color c);
    
    void DrawEllipse (const point& p, coord a, coord b, GsmpColor::Color c);
    void DrawFilledEllipse (const point& p, coord a, coord b, GsmpColor::Color c);
    void DrawAaEllipse (const point& p, coord a, coord b, GsmpColor::Color c);
    
    // TODO: splines or/and Aa curves helper
    
    //protected:
    GsmpColor::Color* m_data;
    
  protected:
    friend class GsmpFace;
    
    // clips the two points to lie in our pixmap area or returns false if it isn't visible
    inline bool InternalClipLine (point& p1, point& p2);
    
    inline void InternalBitBlt (coord x_count, coord y_count, coord per_pixel,
				coord per_src_line, coord per_dest_line,
				GsmpColor::Color* src, GsmpColor::Color* dest);
    
    area m_area;
  };
  
  // ***
  
  class GsmpFace; // forward
  
  class GsmpFontSystem
  {
  public:
    
    struct FontEntry {
      string family;
      string style;
      string filename;
    };
    
    typedef deque<FontEntry>::iterator iterator;
    typedef deque<FontEntry>::const_iterator const_iterator;
    
    GsmpFontSystem ();
    ~GsmpFontSystem ();
    
    // must be called once - initilizes and scans for fonts
    bool Initialize ();
    
    const_iterator Begin () const;
    const_iterator End () const;
    
    GsmpFace* NewFace ();
    
    string GetFileForFace (const string& family,
			   const string& style) const;
    
    void Debug ();
    
  protected:
    
    bool ScanFontDir (const string& dir);
    
    deque<FontEntry> m_fonts;
    
    bool m_initilized;
    FT_Library m_library;
  };
  
  // ***
  
  class GsmpFace
  {
  public:
    GsmpFace (GsmpFontSystem* i_system, FT_Library* i_library);
    ~GsmpFace ();
    
    bool LoadFace (const string& family, const string& style);
    bool UnloadFace ();
    
    /*
     * SetResolution ();
     * SetSize ();
     * DrawText ();
     * DrawText (, int angle);
     */
    
    void Test (GsmpPixmap* t_pixmap);
    
  protected:
    bool m_loaded;
    
    GsmpFontSystem* m_system;
    FT_Library* m_library;
    string m_filename;
    
    FT_Face m_face;
    
    void DrawTfBitmap (GsmpPixmap* t_dest, FT_Bitmap_* t_src,
		       GsmpPixmap::coord dest_x, GsmpPixmap::coord dest_y);
  };
  
} // end namespace GSMP

#endif
